-module(akh_lexer).
-include("akhamoth.hrl").
-export([
    new/1,
    next/1
]).

-define(open_delim(T),
    T =:= $(;
    T =:= $[;
    T =:= ${
).

-define(close_delim(T),
    T =:= $);
    T =:= $];
    T =:= $}
).

-define(op1(T),
    T =:= $,;
    T =:= $;;
    T =:= $.;
    T =:= $+
).

-type token_kind() ::
    comment
    | atom
    | identifier
    | integer
    | '{'
    | '}'
    | '['
    | ']'
    | '('
    | ')'
    | ','
    | ';'
    | ':'
    | '.'
    | '->'
    | '+'.

-define(LOC(Line, Column), {{line, Line}, {column, Column}}).
-type location() :: {{line, pos_integer()}, {column, pos_integer()}} | inserted.

-type token() :: {token_kind(), location()}.

-record(lexer, {
    source :: string(),
    line = 1 :: pos_integer(),
    column = 1 :: pos_integer(),
    delim_stack = [] :: [token()]
}).

-spec new(string()) -> #lexer{}.
new(Source) -> #lexer{source = Source}.

-spec next(#lexer{}) -> none | {ok, token(), #lexer{}}.
next(#lexer{source = [], delim_stack = []}) ->
    none;
next(#lexer{source = [$\n | Rest], line = Line} = Lexer) ->
    next(Lexer#lexer{source = Rest, line = Line + 1, column = 1});
next(#lexer{source = [T | Rest], line = Line, column = Column} = Lexer) when ?op1(T) ->
    Token = {list_to_atom([T]), ?LOC(Line, Column)},
    {ok, Token, Lexer#lexer{source = Rest, column = Column + 1}};
next(#lexer{source = [$-, $> | Rest], line = Line, column = Column} = Lexer) ->
    Token = {'->', ?LOC(Line, Column)},
    {ok, Token, Lexer#lexer{source = Rest, column = Column + 2}};
next(_) ->
    none.
